@@ -4,15 +4,10 @@ use warnings;
 
 BEGIN {
   $Math::Prime::Util::ZetaBigFloat::AUTHORITY = 'cpan:DANAJ';
-  $Math::Prime::Util::ZetaBigFloat::VERSION = '0.44_004';
-}
-
-BEGIN {
-  do { require Math::BigInt;  Math::BigInt->import(try=>"GMP,Pari"); }
-    unless defined $Math::BigInt::VERSION;
-  use Math::BigFloat;
+  $Math::Prime::Util::ZetaBigFloat::VERSION = '0.14';
 }
 
+use Math::BigFloat;
 
 # Riemann Zeta($k) for integer $k.
 # So many terms and digits are used so we can quickly do bignum R.
@@ -202,7 +197,6 @@ my @_Riemann_Zeta_Table = (
 # for k = 1 .. n :  (1 / (zeta(k+1) * k + k)
 # Makes RiemannR run about twice as fast.
 my @_Riemann_Zeta_Premult;
-my $_Riemann_Zeta_Premult_accuracy;
 
 # Select n = 55, good for 46ish digits of accuracy.
 my $_Borwein_n = 55;
@@ -271,53 +265,31 @@ sub _Recompute_Dk {
   $_Borwein_n = $nterms;
   @_Borwein_dk = ();
   foreach my $k (0 .. $nterms) {
+    my $dsum = Math::BigFloat->bzero;
+    $dsum->accuracy(2*$_Borwein_n);
     my $n = Math::BigInt->new($nterms-1)->bfac;
     my $d = Math::BigInt->new($nterms)->bfac;
-    my ($sum_n, $sum_d) = (Math::BigInt->bone, Math::BigInt->bone);
-    my $gcd;
     foreach my $i (0 .. $k) {
-      # ad + cb  /  bd
-      $sum_n->bmul($d)->badd( $sum_d->copy->bmul($n) );
-      $sum_d->bmul($d);
-      $gcd = Math::BigInt::bgcd($sum_n, $sum_d);
-      do {
-        $sum_n = int($sum_n / $gcd);
-        $sum_d = int($sum_d / $gcd);
-      } unless $gcd->is_one;
-      my $dmul = (2*$i+1) * (2*$i+2);
-      $n->bmul($nterms+$i)->blsft(2);
-      $d->bdiv($nterms-$i)->bmul($dmul);
+      my $term = Math::BigFloat->bone;
+      $term->accuracy(2*$_Borwein_n);
+      $term->bmul($n)->bdiv($d);
+      $dsum += $term;
+      $n->bmul($nterms+$i)->bmul(4);
+      $d->bdiv($nterms-$i)->bmul(2*$i+1)->bmul(2*$i+2);
     }
-    $_Borwein_dk[$k] = $sum_n->bmul($nterms)->bdiv($sum_d);
+    my $dk = ($nterms * $dsum + 1e-20)->as_int;
+    $_Borwein_dk[$k] = $dk;
+    #print  "  '$dk',\n";
   }
 }
 
 sub RiemannZeta {
-  my($ix) = @_;
+  my($x, $tol) = @_;
 
-  my $x = (ref($ix) eq 'Math::BigFloat') ? $ix->copy
-                                         : Math::BigFloat->new("$ix");
-  my $xdigits = $x->accuracy || Math::BigFloat->accuracy() || Math::BigFloat->div_scale();
+  return $_Riemann_Zeta_Table[int($x)-2]
+      if $x == int($x) && defined $_Riemann_Zeta_Table[int($x)-2];
 
-  if ($x == int($x) && $xdigits <= 44 && (int($x)-2) <= $#_Riemann_Zeta_Table) {
-    my $izeta = $_Riemann_Zeta_Table[int($x)-2]->copy;
-    $izeta->bround($xdigits);
-    return $izeta;
-  }
-
-  my $extra_acc = 7;
-  if    ($x > 50) { $extra_acc = 10; }
-  elsif ($x > 30) { $extra_acc = 28; }
-  elsif ($x > 15) { $extra_acc = 15; }
-  $xdigits += $extra_acc;
-  $x->accuracy($xdigits);
-  my $zero= $x->copy->bzero;
-  my $one = $x->copy->bone;
-  my $two = $one->copy->binc;
-
-  my $tol = $one->copy->brsft($xdigits-1, 10);
-
-  # Note: with bignum on, $d1->bpow($one-$x) doesn't change d1 !
+  $tol = 1e-40 unless defined $tol;
 
   # Trying to work around Math::BigFloat bugs RT 43692 and RT 77105 which make
   # a right mess of things.  Watch this:
@@ -326,29 +298,26 @@ sub RiemannZeta {
   # We can fix some issues with large exponents (e.g. 6^-40.5) by turning it
   # into (6^-(40.5/4))^4  (assuming the base is positive).  Without that hack,
   # none of this would work at all.
-  # There is a fix for the defect in the RT.
 
-  my $superx = Math::BigInt->bone;
+  $x = Math::BigFloat->new("$x");
+  my $superx = 1;
   my $subx = $x->copy;
-  while ($subx > 1) {
-    $superx->blsft(1);
-    $subx /= $two;
+  while ($subx > 8) {
+    $superx *= 2;
+    $subx /= 2;
   }
 
   # Go with the basic formula for large x, as it best works around the mess,
   # though is unfortunately much slower.
-  if ($x > 50) {
-    my $negsubx = $subx->copy->bneg;
-    my $sum = $zero->copy;
-    my $k = $two->copy->binc;
-    while ($k->binc <= 1000) {
-      my $term = $k->copy->bpow($negsubx)->bpow($superx);
-      $sum->badd($term);
+  if ($x > 30) {
+    my $sum = 0.0;
+    for my $k (4 .. 1000) {
+      my $term = ( $k ** -$subx )  ** $superx;
+      $sum += $term;
       last if $term < ($sum*$tol);
     }
-    $sum->badd( $two->copy->binc->bpow($negsubx)->bpow($superx) );
-    $sum->badd( $two->copy      ->bpow($negsubx)->bpow($superx) );
-    $sum->bround($xdigits-$extra_acc);
+    $sum += ( 3 ** -$subx )  ** $superx;
+    $sum += ( 2 ** -$subx )  ** $superx;
     return $sum;
   }
   #if ($x > 25) {
@@ -366,109 +335,78 @@ sub RiemannZeta {
   #  return $sum;
   #}
 
-  {
-    my $dig = int($_Borwein_n / 1.3)+1;
-    _Recompute_Dk( int($xdigits * 1.3) + 4 )  if $dig < $xdigits;
-  }
+  # If we wanted to change the Borwein series being used:
+  # _Recompute_Dk(55);
 
   if (ref $_Borwein_dk[0] ne 'Math::BigInt') {
     @_Borwein_dk = map { Math::BigInt->new("$_") } @_Borwein_dk;
   }
 
   my $n = $_Borwein_n;
+  my $intermediate_accuracy = undef;
+  my $one = Math::BigFloat->bone;
+  $one->accuracy($intermediate_accuracy) if defined $intermediate_accuracy;
 
-  my $d1 = $two ** ($one - $x);
+  my $d1 = Math::BigFloat->new(2);
+  $d1->accuracy($intermediate_accuracy) if defined $intermediate_accuracy;
+  # with bignum on, $d1->bpow($one-$x) doesn't change d1 !
+  $d1 = $d1 ** ($one - $x);
   my $divisor = $one->copy->bsub($d1)->bmul(-$_Borwein_dk[$n]);
+
   $tol = $divisor->copy->bmul($tol)->babs();
 
-  my ($sum, $bigk) = ($zero->copy, $one->copy);
+  my $sum = Math::BigFloat->bzero;
+  $sum->accuracy($intermediate_accuracy) if defined $intermediate_accuracy;
   foreach my $k (1 .. $n-1) {
-    my $den = $bigk->binc()->copy->bpow($subx,$xdigits)->bpow($superx,$xdigits);
-    my $term = ($k % 2)
-             ? $zero->copy->badd($_Borwein_dk[$n])->bsub($_Borwein_dk[$k])
-             : $zero->copy->badd($_Borwein_dk[$k])->bsub($_Borwein_dk[$n]);
-    $term->bdiv($den);
-    $sum->badd($term);
+    my $term = Math::BigFloat->new( $_Borwein_dk[$k] - $_Borwein_dk[$n] );
+    $term *= -1 if $k % 2;
+    $term->accuracy($intermediate_accuracy) if defined $intermediate_accuracy;
+    my $den = Math::BigFloat->new($k+1);
+    $den->accuracy($intermediate_accuracy) if defined $intermediate_accuracy;
+    $den = ($den ** $subx) ** $superx;
+    $term /= $den;
+    $sum += $term;
     last if $term->copy->babs() < $tol;
   }
-  $sum->badd($one->copy->bsub($_Borwein_dk[$n]));  # term k=0
-  $sum->bdiv($divisor,$xdigits)->bdec;
-  $sum->bround($xdigits-$extra_acc);
+  $sum += Math::BigFloat->new( $one - $_Borwein_dk[$n] ); # term k=0
+  $sum->bdiv( $divisor );
+  $sum->bsub(1);
   return $sum;
 }
 
 # Riemann R function
 sub RiemannR {
-  my($x) = @_;
+  my($x, $tol) = @_;
 
-  if (ref($x) eq 'Math::BigInt') {
-    my $xacc = $x->accuracy();
-    $x = Math::BigFloat->new($x);
-    $x->accuracy($xacc) if $xacc;
+  if (scalar @_Riemann_Zeta_Premult == 0) {
+    @_Riemann_Zeta_Premult = map { my $v = Math::BigFloat->bone;
+                                   $v->accuracy($x->accuracy() || 45);
+                                   $v / ($_Riemann_Zeta_Table[$_-1] * $_ + $_) }
+                             (1 .. @_Riemann_Zeta_Table);
   }
-  $x = Math::BigFloat->new("$x") if ref($x) ne 'Math::BigFloat';
-  my $xdigits = $x->accuracy || Math::BigFloat->accuracy() || Math::BigFloat->div_scale();
-  my $extra_acc = 1;
-  $xdigits += $extra_acc;
-  $x->accuracy($xdigits);
-  my $tol = $x->copy->bone->brsft($xdigits-1, 10);
-  my $sum = $x->copy->bone;
 
-  if ($xdigits <= length($x->copy->as_int->bstr())) {
+  $x = new Math::BigFloat "$x"  if ref($x) ne 'Math::BigFloat';
 
-    for my $k (1 .. 1000) {
-      my $mob = Math::Prime::Util::moebius($k);
-      next if $mob == 0;
-      $mob = Math::BigFloat->new($mob);  $mob->accuracy($xdigits);
-      my $term = $mob->bdiv($k) *
-                 Math::Prime::Util::LogarithmicIntegral($x->copy->broot($k));
-      $sum += $term;
-      #warn "k = $k  term = $term  sum = $sum\n";
-      last if abs($term) < ($tol * abs($sum));
+  $tol = 1e-35 unless defined $tol;
+  my $sum = Math::BigFloat->bone;
+
+  my $flogx = log($x);
+  my $part_term = Math::BigFloat->bone;
+
+  for my $k (1 .. 10000) {
+    my $zeta_term = $_Riemann_Zeta_Premult[$k-1];
+    if (!defined $zeta_term) {
+      my $zeta = (($k-1) <= $#_Riemann_Zeta_Table)
+                 ? $_Riemann_Zeta_Table[$k-1]
+                 : RiemannZeta( $k+1 );
+      $zeta_term = Math::BigFloat->bone / ($zeta * $k + $k);
     }
-
-  } else {
-
-    # TODO: The default table is only 44 digits.
-    if ( (scalar @_Riemann_Zeta_Premult == 0) || ($_Riemann_Zeta_Premult_accuracy < $xdigits) ) {
-      $_Riemann_Zeta_Premult_accuracy = $xdigits;
-      @_Riemann_Zeta_Premult = map { my $v = Math::BigFloat->bone;
-                                     $v->accuracy($xdigits);
-                                     $v / ($_Riemann_Zeta_Table[$_-1]*$_ + $_) }
-                               (1 .. @_Riemann_Zeta_Table);
-    }
-
-    my $flogx = log($x);
-    my $part_term = Math::BigFloat->bone;
-
-    for my $k (1 .. 10000) {
-      my $zeta_term = $_Riemann_Zeta_Premult[$k-1];
-      if (!defined $zeta_term) {
-        my $zeta = $_Riemann_Zeta_Table[$k-1];
-        if (!defined $zeta) {
-          my $kz = Math::BigFloat->new($k+1);
-          $kz->accuracy($xdigits);
-          if ($kz >= 100 && $xdigits <= 40) {
-            # For this accuracy level, two terms are more than enough.  Also,
-            # we should be able to miss the Math::BigFloat accuracy bug.  If we
-            # try to do this for higher accuracy, things will go very bad.
-            $zeta = Math::BigFloat->new(3)->bpow(-$kz)
-                  + Math::BigFloat->new(2)->bpow(-$kz);
-          } else {
-            $zeta = Math::Prime::Util::ZetaBigFloat::RiemannZeta( $kz );
-          }
-        }
-        $zeta_term = Math::BigFloat->bone / ($zeta * $k + $k);
-      }
-      $part_term *= $flogx / $k;
-      my $term = $part_term * $zeta_term;
-      $sum += $term;
-      #warn "k = $k  term = $term  sum = $sum\n";
-      last if $term < ($tol*$sum);
-    }
-
+    $part_term *= $flogx / $k;
+    my $term = $part_term * $zeta_term;
+    #warn "k = $k  term = $term\n";
+    $sum += $term;
+    last if $term < ($tol*$sum);
   }
-  $sum->bround($xdigits-$extra_acc);
   return $sum;
 }
 
@@ -491,33 +429,19 @@ Math::Prime::Util::ZetaBigFloat - Perl Big Float versions of Riemann Zeta and R
 
 =head1 VERSION
 
-Version 0.44_004
+Version 0.14
 
 
 =head1 SYNOPSIS
 
 Math::BigFloat versions`of the Riemann Zeta and Riemann R functions.  These
 are kept in a separate module because they use a lot of big tables that we'd
-prefer to only load if needed.
+prefer not to have loaded all the time.
 
 
 =head1 DESCRIPTION
 
 Pure Perl implementations of Riemann Zeta and Riemann R using Math::BigFloat.
-These functions are used if:
-
-=over 4
-
-=item The input is a BigInt, a BigFloat, or the bignum module has been loaded.
-
-=item The Math::MPFR module is not available.
-
-=back
-
-If you use these functions a lot, I B<highly> recommend you install
-L<Math::MPFR>, which the main L<Math::Prime::Util> functions will find.
-These give B<much> better performance, and better accuracy.  You can also
-use L<Math::Pari> for the Riemann Zeta function.
 
 
 =head1 FUNCTIONS
@@ -531,9 +455,10 @@ point value of ζ(s)-1, where ζ(s) is the Riemann zeta function.  One is
 subtracted to ensure maximum precision for large values of C<s>.  The zeta
 function is the sum from k=1 to infinity of C<1 / k^s>
 
-Results are calculated using either Borwein (1991) algorithm 2, or the basic
-series.  Full input accuracy is attempted, but there are defects in
-Math::BigFloat with high accuracy computations that make this difficult.
+Accuracy should be at least 14 digits, but currently does not increase
+accuracy with big floats.  Small integer values are returned from a table,
+values between 0.5 and 5 use rational Chebyshev approximation, and larger
+values use a series.
 
 
 =head2 RiemannR
@@ -544,29 +469,29 @@ Given a positive non-zero floating point input, returns the floating
 point value of Riemann's R function.  Riemann's R function gives a very close
 approximation to the prime counting function.
 
-Accuracy should be about 35 digits.
+Accuracy should be at least 14 digits.
 
 
 =head1 LIMITATIONS
 
 Bugs in Math::BigFloat (RT 43692, RT 77105) cause many problems with this code.
-I've attempted to work around them, but it is possible there are cases they
+I've attempted to apply workarounds, but it is possible there are cases they
 miss.
 
-The accuracy goals (35 digits) are sometimes missed by a digit or two.
+The accuracy goals (35 digits) are sometimes missed by a digit or two, and
+extensive testing needs to be done to ensure we meet the goals.
 
 
 =head1 PERFORMANCE
 
-Performance is quite bad.
+Performance is not good at all.  A version using XS+GMP would be good to have.
+Pari can give better accuracy in a miniscule fraction of the time.
 
 
 =head1 SEE ALSO
 
 L<Math::Prime::Util>
 
-L<Math::MPFR>
-
 L<Math::Pari>
 
 
